import pathlib
import math
from typing import Dict
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset
import lmdb
import pickle
from torchvision.utils import make_grid
import numpy as np
from einops import rearrange, asnumpy
from collections import namedtuple
from imageio import imread
from cytoolz.dicttoolz import merge_with
from PIL import Image
import nibabel as nib
import torchio


CodeRow = namedtuple("CodeRow", ["filename", "label", "top", "bottom"])
CodeRowDeID = namedtuple("CodeRowDeID", ["filename", "brain_mask", "mri"])


class LMDBDataset(Dataset):
    def __init__(self, path: pathlib.Path):
        self.path = path
        self.txn = None

        # Metadata
        env = lmdb.open(str(self.path), readonly=True, create=False, readahead=False, meminit=False)

        _attrs = [
            "length",
            "num_channels",
            "num_embedding_channels_bottom",
            "num_embedding_channels_top",
            "img_size_bottom",
            "img_size_top",
            "num_codebook_vectors_bottom",
            "num_codebook_vectors_top",
            "num_codebook_vectors_top_mri",
            "num_codebook_vectors_top_brain_mask",
        ]
        with env.begin(write=False) as txn:
            get_int_value = lambda attr: int(
                txn.get(attr.encode("utf-8"), default="-1".encode("utf-8")).decode("utf-8")
            )
            for attr in _attrs:
                setattr(self, attr, get_int_value(attr))

        env.close()

    def open_lmdb(self):
        self.env = lmdb.open(str(self.path), readonly=True, create=False)
        self.txn = self.env.begin(buffers=True)

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        if self.txn is None:
            self.open_lmdb()

        key = str(index).encode("utf-8")
        row = pickle.loads(self.txn.get(key))

        return {
            k: torch.from_numpy(v) if isinstance(v, np.ndarray) else v
            for k, v in row._asdict().items()
            if v is not None
        }


class MriDatasetTorchIO(torchio.SubjectsDataset):
    def __init__(
        self,
        dataset_dir: pathlib.Path,
        index_fname: pathlib.Path,
        data_mode: str,
        img_size: int,
        augmentation_kwargs: Dict,
    ):

        self.dataset_dir = dataset_dir
        self.index_fname = index_fname
        self.img_size = img_size
        self.augmentation_kwargs = augmentation_kwargs

        self.ix = self.index_fname.read_text().strip().split("\n")
        self.fnames = [
            fname for fname in self.dataset_dir.glob("**/*.nii.gz") if fname.name in self.ix
        ]

        subjects = [
            torchio.Subject(
                voxel_values_unaug=torchio.ScalarImage(fname),
                voxel_values_aug=torchio.ScalarImage(fname),
            )
            for fname in self.fnames
        ]

        if data_mode == "mri":
            transform_list = [
                getattr(torchio, augmentation_k)(**augmentation_v, exclude="voxel_values_unaug")
                for augmentation_k, augmentation_v in augmentation_kwargs.items()
                if augmentation_v.get("p", 1.0) > 0.0
            ]
            transform = torchio.Compose(transform_list)
        elif data_mode == "brain_mask":
            transform = torchio.Lambda(function=lambda x: (x != 0).float())
        elif data_mode == "continuous_brain":
            transform = None
        else:
            raise NotImplementedError(f"Unknown data mode {data_mode}")

        super().__init__(subjects=subjects, transform=transform)


class MriDataLoaderTorchIO(DataLoader):
    def __init__(
        self,
        dataset_dir: pathlib.Path,
        index_fname: pathlib.Path,
        prefix: str,
        variant: str,
        data_mode: str,
        img_size: int,
        augmentation_kwargs: Dict,
        batch_size: int,
        num_workers: int,
    ):

        dataset_name = f"{prefix}_{img_size}_{variant}"

        if data_mode == "brain_mask":
            dataset_name += "_brain_mask"
        elif data_mode == "continuous_brain":
            dataset_name += "_brain_mask_fcm_normalized"

        self.dataset_dir = dataset_dir.joinpath(dataset_name)
        assert self.dataset_dir.exists(), f"Dataset directory {self.dataset_dir} does not exist!"

        dataset = MriDatasetTorchIO(
            dataset_dir=self.dataset_dir,
            index_fname=index_fname,
            data_mode=data_mode,
            img_size=img_size,
            augmentation_kwargs=augmentation_kwargs,
        )

        super().__init__(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            pin_memory=True,
        )


class MriDatasetNibabel(Dataset):
    def __init__(self, dataset_dir: pathlib.Path, img_size: int):
        super().__init__()

        self.dataset_dir = dataset_dir
        self.img_size = img_size

        self.fnames = list(self.dataset_dir.glob("**/*.nii.gz"))

    def __getitem__(self, i):
        fname = self.fnames[i]
        relative_fname = str(fname.relative_to(self.dataset_dir))
        label = relative_fname.split("/")[0]
        im = nib.load(fname).get_fdata(dtype=np.float32)
        assert (
            im.shape[0] == self.img_size and im.shape[1] == self.img_size
        ), f"Inconsistent image size for image {relative_fname}"
        im = (im - im.min()) / (im.max() - im.min())
        arr = rearrange(im, "h w d -> 1 h w d")
        return {"voxel_values": torch.from_numpy(arr), "label": label}

    def __len__(self):
        return len(self.fnames)


class MriDataLoaderNibabel(DataLoader):
    def __init__(
        self,
        dataset_dir: pathlib.Path,
        prefix: str,
        img_size: int,
        batch_size: int,
        num_workers: int,
    ):

        self.dataset_dir = dataset_dir.joinpath(f"{prefix}_{img_size}")
        assert self.dataset_dir.exists(), f"Dataset directory {self.dataset_dir} does not exist!"
        dataset = MriDatasetNibabel(dataset_dir=self.dataset_dir, img_size=img_size)

        super().__init__(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            pin_memory=True,
        )


class ImageDataLoader(DataLoader):
    def __init__(self, dataset_dir: pathlib.Path, img_size: int, batch_size: int, num_workers: int):

        dataset = ImageDataset(dataset_dir=dataset_dir, img_size=img_size)
        self.num_color_channels = dataset.num_color_channels

        super().__init__(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            pin_memory=True,
            collate_fn=self._collate_fn,
        )

    def _collate_fn(self, xs):
        batch = merge_with(list, *xs)
        batch["pixel_values"] = torch.stack(batch["pixel_values"], dim=0)
        batch["pixel_values"] = 2 * batch["pixel_values"] - 1
        return batch


class ImageDataset(Dataset):
    def __init__(self, dataset_dir: pathlib.Path, img_size: int):
        super().__init__()

        self.dataset_dir = dataset_dir
        self.img_size = img_size

        self.fnames = list(self.dataset_dir.glob("**/*.jpg"))
        self.num_color_channels = None
        with Image.open(self.fnames[0]) as im:
            self.num_color_channels = len(im.mode)

    def __getitem__(self, i):
        fname = self.fnames[i]
        relative_fname = str(fname.relative_to(self.dataset_dir))
        label = relative_fname.split("/")[0]
        im = imread(fname)
        assert (
            im.shape[0] == self.img_size and im.shape[1] == self.img_size
        ), f"Inconsistent image size for image {relative_fname}"
        arr = rearrange(np.asarray(im), "h w c -> c h w")
        arr = arr.astype(np.float32) / 255.0
        return {"pixel_values": torch.from_numpy(arr), "label": label, "fname": relative_fname}

    def __len__(self):
        return len(self.fnames)


def tensor_to_image_grid(tensor: torch.FloatTensor, shift: bool, nrow: int = 8):
    if shift:
        tensor = torch.clamp((tensor + 1) / 2, 0.0, 1.0)

    grid = asnumpy(rearrange(make_grid(tensor, nrow=nrow), "c h w -> h w c"))

    return grid


def expand_tensor_to_4d(tensor: torch.Tensor):
    return rearrange(tensor, "b -> b 1 1 1")
