"""
Uses https://github.com/lucidrains/bit-diffusion as a base which itself uses a categorical diffusion scenario.
This code suggests a couple of improvements and adapts the scenario for masked autodecoding and de-identification.
"""
import torch
from einops import rearrange, einsum, repeat, reduce
from torch import nn as nn
from torch.nn import functional as F

from utils import exists


def uniform(*shape, a: float, b: float, device: torch.device):
    return (b - a) * torch.rand(*shape, device=device) + a


class CpMae(nn.Module):
    def __init__(
            self,
            model,
            *,
            img_size: int,
            num_states: int,
            num_states_condition: int,
            num_timesteps: int,
            noise_schedule: str,
            label_smoothing: float,
            eps: float,
            use_temperature_annealing: bool,
            num_renoising_steps: bool
    ):
        super().__init__()
        self.model = model
        self.img_size = img_size
        self.num_states = num_states
        self.num_states_condition = num_states_condition
        self.num_timesteps = num_timesteps
        self.mask_schedule = noise_schedule
        self.label_smoothing = label_smoothing
        self.eps = eps
        self.use_temperature_annealing = use_temperature_annealing
        self.num_renoising_steps = num_renoising_steps

        self.num_img_tokens = self.img_size ** 3

    @property
    def device(self):
        return next(self.model.parameters()).device

    def apply_noise_schedule(self, t: torch.FloatTensor):
        if self.mask_schedule == 'linear':
            return linear_beta_schedule(t)
        elif self.mask_schedule == 'quadratic':
            return quadratic_beta_schedule(t)
        elif self.mask_schedule == 'cubic':
            return cubic_beta_schedule(t)
        elif self.mask_schedule == 'cosine':
            return cosine_beta_schedule(t)
        elif self.mask_schedule == 'sigmoid':
            return sigmoid_beta_schedule(t)
        else:
            raise NotImplementedError('Unknown noise schedule!')

    def q_sample(self, x_start, t, x_random: torch.LongTensor = None):
        noise_probs = repeat(self.apply_noise_schedule(t), 'n -> n h w d 1', h=self.img_size, w=self.img_size, d=self.img_size)
        noise_mask = torch.bernoulli(noise_probs).type_as(x_start)

        if x_random is None:
            x_random = torch.randint_like(x_start, 0, self.num_states)

        x_noisy = (1 - noise_mask) * x_start + noise_mask * x_random
        return x_noisy

    @torch.inference_mode()
    def sample(self, condition: torch.LongTensor, x_init=None, mask=None):
        return self.p_sample_loop(shape=(len(condition), self.img_size, self.img_size, self.img_size, 1),
                                  condition=rearrange(condition, '... -> ... 1'),
                                  x_init=x_init,
                                  mask=mask)

    @torch.inference_mode()
    def p_sample_loop(self, shape, condition: torch.LongTensor, x_init: torch.LongTensor = None, mask: torch.BoolTensor = None):
        """
        :param shape: Of shape [N, H, W, 1]
        :param mask: Boolean of shape [H, W, 1]. True where elements should be filled
        """
        assert exists(x_init) == exists(mask), "Mask needs to exist when x_init is given!"
        assert not exists(x_init) or x_init.shape == shape, "Tensor 'x_init' has wrong shape!"
        assert not exists(mask) or mask.shape == shape[1:], "Mask tensor has wrong shape!"
        assert not exists(x_init) or not exists(mask) or torch.all(x_init[:, mask] == -1), "Information was not properly erased!"

        ts = torch.linspace(0.0, 1.0, self.num_timesteps + 1, device=self.device)[1:]

        if not exists(x_init):
            x_init = torch.full(size=shape, fill_value=self.num_states, device=self.device, dtype=torch.int64)

        if not exists(mask):
            mask = torch.ones(*shape[1:], dtype=torch.bool, device=self.device)

        x_init = torch.where(mask, torch.randint(0, self.num_states, shape, device=self.device, dtype=torch.int64), x_init)

        if self.use_temperature_annealing:
            temperature = torch.linspace(1.0, 0.0, self.num_timesteps + 1, device=self.device)[:-1]
        else:
            temperature = torch.ones(self.num_timesteps, device=self.device)

        def p_sample(i, x, is_last_iter):

            x_prev = x.clone()
            t_prev = repeat(ts[i], ' -> n', n=x.shape[0])

            # implicitly assumes that x_prev = self.q_sample(x_start, t_prev)
            x_next_logits = compatibility_wrapper(self.model,
                                                  x=x_prev,
                                                  t=t_prev,
                                                  condition=condition,
                                                  num_states=self.num_states,
                                                  num_states_condition=self.num_states_condition,
                                                  is_training=False)
            x_next_probs = F.softmax(x_next_logits/temperature[i], dim=-1)
            x_next = sample_from_categorical(probs=x_next_probs, deterministic=is_last_iter)

            if not is_last_iter and i < self.num_renoising_steps:
                t_next = repeat(ts[i + 1], ' -> n', n=x.shape[0])
                x_next = self.q_sample(x_next, t=t_next, x_random=x_init)

            x_next[:, ~mask] = x_init[:, ~mask]

            return x_next

        # High noise -> low noise
        x_final = x_init.clone()
        for i in range(self.num_timesteps):
            x_final = p_sample(i=i, x=x_final, is_last_iter=(i == self.num_timesteps - 1))

        return x_final  # [N, H, W, D, 1]

    def forward(self,
                x_start: torch.LongTensor,  # [N, H, W, D]
                condition: torch.LongTensor # [N, H, W, D]
                ):
        N = len(x_start)

        x_start = rearrange(x_start, "... -> ... 1")  # [N, H, W, D, 1]
        condition = rearrange(condition, '... -> ... 1')   # [N, H, W, D, 1]

        t = torch.rand(size=(N,), device=self.device, dtype=torch.float32)  # [N]
        x_t = self.q_sample(x_start=x_start, t=t)  # [N, H, W, 1]

        x_start_pred = compatibility_wrapper(self.model,
                                             x=x_t,
                                             t=t,
                                             condition=condition,
                                             num_states=self.num_states,
                                             num_states_condition=self.num_states_condition,
                                             is_training=True)
        loss_dict = {'mvtm': F.cross_entropy(input=rearrange(x_start_pred, "n ... c -> n c ..."), target=x_start, label_smoothing=self.label_smoothing)}

        loss_dict['loss'] = loss_dict['mvtm']
        return loss_dict, x_t[..., 0]


def sample_from_categorical(probs: torch.FloatTensor, deterministic: bool = False):
    noise = torch.rand_like(probs)
    noise = torch.clamp(noise, min=torch.finfo(noise.dtype).tiny, max=1.0)
    gumbel_noise = -torch.log(-torch.log(noise))

    if deterministic:
        mask = torch.zeros_like(gumbel_noise)
    else:
        mask = torch.ones_like(gumbel_noise)

    sample = torch.argmax(torch.log(probs) + mask * gumbel_noise, dim=-1)
    return sample


def linear_beta_schedule(t: torch.FloatTensor):
    return 1 - t


def quadratic_beta_schedule(t: torch.FloatTensor):
    return 1 - t ** 2


def cubic_beta_schedule(t: torch.FloatTensor):
    return 1 - t ** 3


def cosine_beta_schedule(t: torch.FloatTensor):
    return torch.cos(t * torch.pi / 2.)


def sigmoid_beta_schedule(t: torch.FloatTensor, start=-3, end=3, tau=1, clamp_min=1e-5):
    """
    sigmoid schedule
    proposed in https://arxiv.org/abs/2212.11972 - Figure 8
    better for images > 64x64, when used during training
    """
    v_start = torch.tensor(start / tau).sigmoid()
    v_end = torch.tensor(end / tau).sigmoid()
    output = (-torch.sigmoid((t * (end - start) + start) / tau) + v_end) / (v_end - v_start)
    return torch.clamp(output, clamp_min, 1.)


def compatibility_wrapper(
        model_fn,
        x: torch.LongTensor,  # [N, H, W, D, 1]
        t: torch.LongTensor,  # [N]
        condition: torch.LongTensor,  # [N, H, W, D, 1]
        num_states: int,
        num_states_condition: int,
        is_training: bool,
        self_critic_mode: bool = False
):
    x = x[..., 0]  # [N, H, W, D]
    x = 2.0 * F.one_hot(x, num_classes=num_states) - 1.0  # [N, H, W, D, #CV]
    x = nhwdc_to_nchwd(x)  # [N, #CV, H, W, D]
    x = x.contiguous()

    condition = condition[..., 0]   # [N, H, W, D]
    condition = 2.0 * F.one_hot(condition, num_classes=num_states_condition) - 1.0    # [N, H, W, D, #CV]
    condition = nhwdc_to_nchwd(condition)  # [N, #CV, H, W, D]
    x = model_fn(x=x,
                 noise_level=t,
                 condition=condition,
                 return_pre_logits=self_critic_mode)
    x = rearrange(x, "n cv h w d -> n h w d 1 cv")  # [N, H, W, D, 1, #CV]

    return x


def sum_over_non_batch_dims(x):
    """Take the sum over all axes except the first batch dimension."""
    return x.sum(dim=tuple(range(1, len(x.shape))))


def numerically_stable_log(t, eps):
    return torch.log(t + eps)


def nhwdc_to_nchwd(x):
    return rearrange(x, "n h w d c -> n c h w d")


def nchwd_to_nhwdc(x):
    return rearrange(x, "n c h w d -> n h w d c")


def to_decimal(tensor: torch.FloatTensor, num_states: int):
    return tensor[..., 0]
