import importlib
import os
from argparse import ArgumentParser
import pathlib
import random

import numpy as np
import randomname
import oyaml as yaml

import torch
import torch.nn.functional as F
import wandb
from accelerate import Accelerator, DistributedDataParallelKwargs
from cytoolz import keyfilter, valmap, keymap
from easydict import EasyDict
from ema_pytorch import EMA
from psutil import cpu_count
from torch.utils.data import DataLoader
from tqdm import tqdm
from einops import rearrange, repeat, asnumpy
from torchinfo import summary

from discrete_synthesis.cp_mae import CpMae, to_decimal
from data import LMDBDataset, tensor_to_image_grid
from vqa.einops_vqa import Autoencoder

from lion_pytorch import Lion
from torch.optim import Adam
from utils import visualize_mri_slices, strip_distributed_markers

from einops._torch_specific import allow_ops_in_compiled_graph  # requires einops>=0.6.1

allow_ops_in_compiled_graph()


def seed_everything(seed: int):
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = False
    torch.backends.cudnn.benchmark = True


def pick(d, keys):
    return keyfilter(lambda k: k in keys, d)


def to_scaled_ohe(tensor: torch.LongTensor, num_states: int):
    return 2.0 * F.one_hot(tensor, num_classes=num_states) - 1.0


def exists(x):
    return x is not None


def cycle(dl):
    while True:
        for data in dl:
            yield data


def default(val, d):
    if exists(val):
        return val
    return d() if callable(d) else d


def append_dim(tensor):
    return rearrange(tensor, "... -> ... 1")


def main(lmdb_dir: pathlib.Path, cp_mae_config_fname: pathlib.Path, autoencoder_mri_config_fname: pathlib.Path, autoencoder_mri_ckpt_fname: pathlib.Path, dry_run: bool, load_ckpt: bool):
    # Config
    with open(cp_mae_config_fname, "r") as config_f:
        cp_mae_config = EasyDict(yaml.load(config_f, Loader=yaml.FullLoader))

    with open(autoencoder_mri_config_fname, "r") as config_f:
        autoencoder_mri_config = EasyDict(yaml.load(config_f, Loader=yaml.FullLoader))

    ddp_kwargs_top = DistributedDataParallelKwargs(find_unused_parameters=cp_mae_config.mixed_precision != "no" and cp_mae_config.cp_mae_top.is_self_conditional)
    accelerator_top = Accelerator(mixed_precision=cp_mae_config.mixed_precision, kwargs_handlers=[ddp_kwargs_top], gradient_accumulation_steps=cp_mae_config.num_accumulation_steps)
    is_local_main_process = accelerator_top.is_local_main_process

    tmp_dir = autoencoder_mri_ckpt_fname.parent
    cp_mae_ckpt_fname = tmp_dir.joinpath(f"cp_mae_{cp_mae_config.model}.pt")

    if is_local_main_process and not dry_run:
        wandb.init(project="<PROJECT>", entity="<USER>", name=f"{cp_mae_config.model}-{randomname.get_name()}")
        wandb.run.log_code(".")
        wandb.define_metric("train/num_gradient_updates")
        wandb.define_metric("train/*", step_metric="train/num_gradient_updates")

    if is_local_main_process and not dry_run:
        wandb.config.update({"cp_mae": cp_mae_config, "autoencoder": autoencoder_mri_config})

    dataset = LMDBDataset(path=lmdb_dir)
    assert autoencoder_mri_config.data_mode == "mri", "Wrong data mode, change config!"
    data = DataLoader(dataset=dataset, batch_size=cp_mae_config.batch_size, shuffle=True, pin_memory=True, num_workers=cpu_count())

    if is_local_main_process and not dry_run:
        with open(cp_mae_config_fname, "r") as f_in, open(tmp_dir.joinpath("config_cp_mae.yml"), "w") as f_out:
            _cp_mae_config = yaml.load(f_in, Loader=yaml.FullLoader)
            _cp_mae_config["img_size_top"] = dataset.img_size_top
            _cp_mae_config["num_codebook_vectors_top_mri"] = dataset.num_codebook_vectors_top_mri
            _cp_mae_config["num_channels"] = dataset.num_channels

            yaml.dump(_cp_mae_config, f_out, default_flow_style=False)

    ModelCls = getattr(importlib.import_module(f"architectures.masked_decoding.{cp_mae_config.model}"), "Model")
    model_top = ModelCls(
        img_size=dataset.img_size_top,
        num_input_channels=dataset.num_codebook_vectors_top_mri,
        num_output_channels=dataset.num_codebook_vectors_top_mri,
        num_input_channels_condition=dataset.num_codebook_vectors_top_brain_mask,
        is_self_conditional=False,
        is_conditional=True,
        is_training=True,
        **cp_mae_config.model_top,
    ).to(device=accelerator_top.device)


    cp_mae_top = CpMae(
        model=model_top,
        img_size=dataset.img_size_top,
        num_states=dataset.num_codebook_vectors_top_mri,
        num_states_condition=dataset.num_codebook_vectors_top_brain_mask,
        **cp_mae_config.generative_top,
    )

    if is_local_main_process:
        autoencoder_mri = (
            Autoencoder(
                num_input_channels=1,
                img_size=autoencoder_mri_config.img_size,
                img_size_top=autoencoder_mri_config.img_size_top,
                is_training=False,
                **autoencoder_mri_config.autoencoder,
            )
            .to(device=accelerator_top.device)
            .eval()
        )
        autoencoder_mri = torch.compile(autoencoder_mri, backend="inductor")

        ckpt = torch.load(autoencoder_mri_ckpt_fname, map_location="cpu")
        stripped_ckpt = strip_distributed_markers(ckpt, pattern="module.", replacement="")
        autoencoder_mri.load_state_dict(stripped_ckpt)

    OptimizerClsTop = Lion if cp_mae_config.optimizer_top.use_lion else Adam
    optimizer_top = OptimizerClsTop(cp_mae_top.parameters(), **pick(cp_mae_config.optimizer_top, ["lr", "betas"]))

    ema_top = None
    if is_local_main_process:
        ema_top = EMA(cp_mae_top, beta=cp_mae_config.ema_decay, update_every=cp_mae_config.ema_step).to(accelerator_top.device)

    cp_mae_top, optimizer_top, data = accelerator_top.prepare(cp_mae_top, optimizer_top, data)

    data = cycle(data)

    if load_ckpt and cp_mae_ckpt_fname.exists():
        state_dict = torch.load(cp_mae_ckpt_fname, map_location="cpu")
        accelerator_top.unwrap_model(cp_mae_top).load_state_dict(state_dict["top"])

    global_step = 0
    total_losses = {}
    for batch in tqdm(data, disable=not is_local_main_process):
        num_gradient_updates = global_step // cp_mae_config.num_accumulation_steps
        local_accumulation_step = global_step % cp_mae_config.num_accumulation_steps

        batch = EasyDict(batch)

        accelerator_top.wait_for_everyone()
        with accelerator_top.accumulate(cp_mae_top):
            with accelerator_top.autocast():
                summary(cp_mae_top, input_data=[batch.mri, batch.brain_mask])
                loss_top_dict, batch.top_noisy = cp_mae_top(batch.mri, condition=batch.brain_mask)

            for k, v in loss_top_dict.items():
                k_ext = f"top_{k}"
                if k not in total_losses:
                    total_losses[k_ext] = np.zeros(shape=[cp_mae_config.num_accumulation_steps])

                total_losses[k_ext][local_accumulation_step] = v.item()

            for i, k in enumerate(cp_mae_config.optimizer_top.backward_on_losses):
                is_last_loss = i == len(cp_mae_config.optimizer_top.backward_on_losses) - 1
                accelerator_top.backward(loss_top_dict[k], retain_graph=not is_last_loss)

            optimizer_top.step()
            optimizer_top.zero_grad()

        if dry_run or not is_local_main_process or num_gradient_updates == 0 or local_accumulation_step != 0:
            global_step += 1
            continue

        ema_top.update()

        if num_gradient_updates % cp_mae_config.save_step == 0:
            print(f"[{num_gradient_updates}]: Saving model weights...")

            ckpt_dict = dict()
            ckpt_dict["top"] = accelerator_top.get_state_dict(accelerator_top.unwrap_model(cp_mae_top))
            ckpt_dict["ema_top"] = ema_top.state_dict()

            # torch.save(ckpt_dict, cp_mae_ckpt_fname)

        with torch.no_grad():
            if num_gradient_updates % cp_mae_config.log_step == 0:
                avg_losses = keymap(lambda k: f"train/{k}", valmap(np.mean, total_losses))

                log_dict = {"train/num_gradient_updates": num_gradient_updates}
                log_dict = {**log_dict, **avg_losses}

                if num_gradient_updates % cp_mae_config.visualization_step == 0:
                    ema_top.ema_model.eval()

                    # Ground truth
                    batch.quantized_top = autoencoder_mri.discrete_to_codebook_vectors(discrete=batch.mri)
                    image_gt = autoencoder_mri.decode_discrete(quantized_top=batch.quantized_top)
                    slices_gt = tensor_to_image_grid(visualize_mri_slices(image_gt, type="axial", min_max_scaling=True), shift=False)

                    # Noisy ground truth
                    batch.top_noisy = torch.clamp(batch.top_noisy, min=0, max=dataset.num_codebook_vectors_top_mri - 1)
                    batch.quantized_top_noisy = autoencoder_mri.discrete_to_codebook_vectors(discrete=batch.top_noisy)
                    image_noisy = autoencoder_mri.decode_discrete(quantized_top=batch.quantized_top_noisy)
                    slices_noisy = tensor_to_image_grid(visualize_mri_slices(image_noisy, type="axial", min_max_scaling=True), shift=False)

                    # Ancestral sampling
                    sampled_top = ema_top.ema_model.sample(condition=batch.brain_mask)
                    decimal_sampled_top = to_decimal(sampled_top, num_states=dataset.num_codebook_vectors_top_mri)
                    quantized_sampled_top = autoencoder_mri.discrete_to_codebook_vectors(discrete=decimal_sampled_top)
                    image_sampled = autoencoder_mri.decode_discrete(quantized_top=quantized_sampled_top)
                    slices_sampled = tensor_to_image_grid(visualize_mri_slices(image_sampled, type="axial", min_max_scaling=True), shift=False)

                    log_dict["train/image_gt"] = wandb.Image(slices_gt)
                    log_dict["train/image_gt_noisy"] = wandb.Image(slices_noisy)
                    log_dict["train/image_pred"] = wandb.Image(slices_sampled)

                    ema_top.ema_model.train()

                wandb.log(log_dict)

        global_step += 1


if __name__ == "__main__":
    seed_everything(42)

    parser = ArgumentParser()

    parser.add_argument("--lmdb_dir", help="Path to the LMDB dataset", required=True)

    parser.add_argument("--cp_mae_config_fname", required=True)

    parser.add_argument("--autoencoder_mri_config_fname", required=True)

    parser.add_argument("--autoencoder_mri_ckpt_fname", required=True)

    parser.add_argument("--dry_run", action="store_true", default=False)

    parser.add_argument("--load_ckpt", action="store_true", default=False)

    args = parser.parse_args()

    args.lmdb_dir = pathlib.Path(args.lmdb_dir)
    assert args.lmdb_dir.exists(), "LMDB dataset does not exist!"

    args.cp_mae_config_fname = pathlib.Path(args.cp_mae_config_fname)
    assert args.cp_mae_config_fname.exists(), "CpMae config does not exist!"

    args.autoencoder_mri_config_fname = pathlib.Path(args.autoencoder_mri_config_fname)
    assert args.autoencoder_mri_config_fname.exists(), "Autoencoder config does not exist!"

    args.autoencoder_mri_ckpt_fname = pathlib.Path(args.autoencoder_mri_ckpt_fname)
    assert args.autoencoder_mri_ckpt_fname.exists(), "Autoencoder checkpoint file does not exist!"

    main(
        lmdb_dir=args.lmdb_dir,
        cp_mae_config_fname=args.cp_mae_config_fname,
        autoencoder_mri_config_fname=args.autoencoder_mri_config_fname,
        autoencoder_mri_ckpt_fname=args.autoencoder_mri_ckpt_fname,
        dry_run=args.dry_run,
        load_ckpt=args.load_ckpt,
    )
