import numpy as np
import pandas as pd
import pathlib
import subprocess
import shlex
import shutil
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.autograd as autograd
from cytoolz.dicttoolz import keyfilter
from einops import rearrange
import plotly.express as px


def exists(x):
    return x is not None


def default(val, d):
    if exists(val):
        return val
    return d() if callable(d) else d


def identity(t, *args, **kwargs):
    return t


def cycle(dl):
    while True:
        for data in dl:
            yield data


def right_pad_dims_to(x, t):
    padding_dims = x.ndim - t.ndim
    if padding_dims <= 0:
        return t
    return t.view(*t.shape, *((1,) * padding_dims))


def append_singleton_dim(t):
    return rearrange(t, '... -> ... 1')


def num_to_groups(num, divisor):
    groups = num // divisor
    remainder = num % divisor
    arr = [divisor] * groups
    if remainder > 0:
        arr.append(remainder)
    return arr


def mean_over_non_batch_dims(x):
    """Take the mean over all axes except the first batch dimension."""
    return x.mean(dim=tuple(range(1, len(x.shape))))


def min_over_non_batch_dims(x, keepdim=False):
    return torch.amin(x, dim=tuple(range(1, len(x.shape))), keepdim=keepdim)


def max_over_non_batch_dims(x, keepdim=False):
    return torch.amax(x, dim=tuple(range(1, len(x.shape))), keepdim=keepdim)


def extract(a, t, x_shape):
    b, *_ = t.shape
    out = a.gather(-1, t)
    return out.reshape(b, *((1,) * (len(x_shape) - 1)))


class STEFunction(autograd.Function):
    @staticmethod
    def forward(ctx, input):
        return (input > 0).float()

    @staticmethod
    def backward(ctx, grad_output):
        return F.hardtanh(grad_output)


class StraightThroughEstimator(nn.Module):
    def __init__(self):
        super(StraightThroughEstimator, self).__init__()

    def forward(self, x):
        x = STEFunction.apply(x)
        return x


def binary_logarithm(n):
    assert is_power_of_two(n), "Input needs to be a power of two!"
    bitstring = bin(n)[2:]
    return len(bitstring) - 1


def is_power_of_two(n: int):
    return n & (n - 1) == 0 and n != 0


def is_square(n: int) -> bool:
    return n == math.isqrt(n) ** 2


def strip_distributed_markers(ckpt, pattern="module", replacement=""):
    from collections import OrderedDict

    return OrderedDict([(k.replace(pattern, replacement), v) for k, v in ckpt.items()])


def pick(d, keys):
    return keyfilter(lambda k: k in keys, d)


def sample_from_probs(probs: torch.FloatTensor, num_samples=1, deterministic: bool = True):
    if deterministic:
        _, indices = torch.topk(probs, k=num_samples, dim=-1)

        return indices
    else:
        distribution = torch.distributions.Categorical(probs=probs)
        return distribution.sample()


def copy_honoring_gitignore(src: pathlib.Path, dst: pathlib.Path):
    assert src.exists(), "Source directory does not exist!"
    assert dst.exists(), "Destination directory does not exist!"

    print("Started copying files..")

    fnames = subprocess.Popen(shlex.split("find . -type d,f -print"), stdout=subprocess.PIPE)
    fnames_ignored = subprocess.check_output(shlex.split("git check-ignore --stdin"), stdin=fnames.stdout).decode().strip().split("\n")
    fnames_ignored = [pathlib.Path(fname).resolve() for fname in fnames_ignored + [".git"]]

    def _copy_func(a: str, b: str):
        fname_a, fname_b = pathlib.Path(a), pathlib.Path(b)
        to_be_ignored = any(fname_ignored in fname_a.parents for fname_ignored in fnames_ignored)

        if not to_be_ignored:
            print(a)
            shutil.copy2(src=a, dst=b)

    shutil.copytree(src=src, dst=dst, dirs_exist_ok=True, copy_function=_copy_func)

    print("Finished copying files!")


def compute_discrete_histogram(discrete_tensor: torch.LongTensor, num_discrete_states: int):
    discrete_arr = discrete_tensor.detach().cpu().numpy().reshape(-1)
    histogram = np.zeros(shape=(num_discrete_states,))
    unique_ids, unique_counts = np.unique(discrete_arr, return_counts=True)
    histogram[unique_ids] = unique_counts
    freq = histogram / histogram.sum()
    embedding_id = np.arange(num_discrete_states, dtype=int)
    return pd.DataFrame({"embedding_id": embedding_id, "freq": freq})


def compute_discrete_histogram_fig(discrete_ids, num_states):
    discrete_histogram = compute_discrete_histogram(discrete_ids, num_discrete_states=num_states)
    fig = px.bar(discrete_histogram, x="embedding_id", y="freq")
    fig.update_layout(bargap=0.0)
    nnz_percentage = 100.0 * (discrete_histogram["freq"] > 0).mean()
    return fig, nnz_percentage


def sequence_mask(lengths, maxlen, dtype=torch.bool):
    """c.f. https://discuss.pytorch.org/t/pytorch-equivalent-for-tf-sequence-mask/39036/2"""
    if maxlen is None:
        maxlen = lengths.max()
    mask = ~(torch.ones((len(lengths), maxlen)).cumsum(dim=1).t() > lengths).t()
    mask.type(dtype)
    return mask


def nchw_to_nhwc(tensor):
    return rearrange(tensor, "n c h w -> n h w c")


def nhwc_to_nchw(tensor):
    return rearrange(tensor, "n h w c -> n c h w")


def visualize_mri_slices(tensor,
                         type: str,
                         slice_locs=(.05, .125, .25, .375, .5, .625, .75, .875, .95),
                         min_max_scaling=False):
    D, H, W = tensor.shape[-3:]
    l = math.isqrt(len(slice_locs))
    assert l * l == len(slice_locs), "Number of location needs to be a square number!"

    tensor = tensor.squeeze(axis=1)

    if min_max_scaling:
        tensor_min = min_over_non_batch_dims(tensor, keepdim=True)
        tensor_max = max_over_non_batch_dims(tensor, keepdim=True)
        tensor = (tensor - tensor_min)/(tensor_max - tensor_min)

    if type == 'sagittal':
        d_slices = [int(slice_loc * D) for slice_loc in slice_locs]
        arr = torch.stack([tensor[:, d_slice, :, :] for d_slice in d_slices], dim=1)  # [N, L, H, W]
        arr = rearrange(arr, 'n (l1 l2) h w -> n 1 (l1 h) (l2 w)', l1=l, l2=l)  # [N, 1, l * H, l * W]
    elif type == 'coronal':
        h_slices = [int(slice_loc * H) for slice_loc in slice_locs]
        arr = torch.stack([tensor[:, :, h_slice, :] for h_slice in h_slices], dim=1)  # [N, L, D, W]
        arr = rearrange(arr, 'n (l1 l2) d w -> n 1 (l1 d) (l2 w)', l1=l, l2=l)  # [N, 1, l * D, l * W]
    elif type == 'axial':
        w_slices = [int(slice_loc * W) for slice_loc in slice_locs]
        arr = torch.stack([tensor[:, :, :, w_slice] for w_slice in w_slices], dim=1)  # [N, L, D, H]
        arr = rearrange(arr, 'n (l1 l2) d h -> n 1 (l1 d) (l2 h)', l1=l, l2=l)  # [N, 1, l * D, l * H]
    else:
        raise ValueError(f'Unknown type {type}')

    return arr
