from functools import partial
import torch.nn as nn
import torch

from utils import binary_logarithm


class LayerNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.g = nn.Parameter(torch.ones(1, dim, 1, 1, 1))

    def forward(self, x):
        assert x.shape[1] > 1, "Need to have more than one channel to perform layer norm!"
        eps = 1e-5 if x.dtype == torch.float32 else 1e-3
        var = torch.var(x, dim=1, unbiased=False, keepdim=True)
        mean = torch.mean(x, dim=1, keepdim=True)
        return (x - mean) * (var + eps).rsqrt() * self.g


class Conv3dLayer(nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size, use_norm, stride=1, padding=0):
        super().__init__(
            LayerNorm(in_channels) if use_norm else nn.Identity(),
            nn.Conv3d(in_channels=in_channels,
                      out_channels=out_channels,
                      kernel_size=kernel_size,
                      stride=stride,
                      padding=padding)
        )

class ConvTranspose3dLayer(nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size, use_norm, stride=1, padding=0):
        super().__init__(
            LayerNorm(in_channels) if use_norm else nn.Identity(),
            nn.ConvTranspose3d(in_channels=in_channels,
                      out_channels=out_channels,
                      kernel_size=kernel_size,
                      stride=stride,
                      padding=padding)
        )

class ResBlock(nn.Module):
    def __init__(self, in_channels, out_channels, activation: str, use_norm):
        super().__init__()

        self.use_norm = use_norm

        ActivationCls = partial(getattr(nn, activation), inplace=True)
        self.conv = nn.Sequential(
            LayerNorm(in_channels) if use_norm else nn.Identity(),
            Conv3dLayer(in_channels, out_channels, 3, padding=1, use_norm=use_norm),
            ActivationCls(),
            LayerNorm(out_channels) if use_norm else nn.Identity(),
            Conv3dLayer(out_channels, in_channels, 1, use_norm=use_norm),
            ActivationCls()
        )

    def forward(self, x):
        return x + self.conv(x)


class Encoder(nn.Module):
    def __init__(self,
                 input_size: int,
                 output_size: int,
                 num_input_channels: int,
                 num_channels: int,
                 num_layers_per_block: int,
                 activation: str,
                 use_norm: bool,
                 **kwargs):
        super().__init__()

        self.input_size = input_size
        self.output_size = output_size
        self.num_input_channels = num_input_channels
        self.num_channels = num_channels
        self.num_downsampling_layers = binary_logarithm(input_size // output_size)
        self.num_layers_per_block = num_layers_per_block

        ActivationCls = partial(getattr(nn, activation), inplace=True)

        blocks = []
        for i in range(self.num_downsampling_layers):
            blocks.extend(
                [
                    nn.Sequential(*[ResBlock(in_channels=self.num_input_channels if i == 0 else self.num_channels,
                                             out_channels=2 * self.num_input_channels if i == 0 else 2 * self.num_channels,
                                             activation=activation,
                                             use_norm=False if i == 0 else use_norm) for _ in range(self.num_layers_per_block)]),
                    Conv3dLayer(in_channels=self.num_input_channels if i == 0 else self.num_channels,
                                out_channels=self.num_channels,
                                kernel_size=4,
                                stride=2,
                                padding=1,
                                use_norm=False),
                    ActivationCls() if i != self.num_downsampling_layers - 1 else nn.Identity(),
                ]
            )

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x):
        return self.blocks(x)


class Decoder(nn.Module):
    def __init__(self,
                 input_size: int,
                 output_size: int,
                 num_output_channels: int,
                 num_channels: int,
                 num_layers_per_block: int,
                 activation: str,
                 use_norm: bool,
                 **kwargs):
        super().__init__()

        self.input_size = input_size
        self.output_size = output_size
        self.num_output_channels = num_output_channels
        self.num_channels = num_channels
        self.num_layers_per_block = num_layers_per_block
        self.num_upsampling_layers = binary_logarithm(input_size // output_size)

        ActivationCls = partial(getattr(nn, activation), inplace=True)

        blocks = []
        for i in range(self.num_upsampling_layers):
            blocks.extend(
                [
                    nn.Sequential(*[ResBlock(in_channels=self.num_channels,
                                             out_channels=2 * self.num_channels,
                                             activation=activation,
                                             use_norm=use_norm) for _ in range(self.num_layers_per_block)]),
                    ConvTranspose3dLayer(
                        in_channels=self.num_channels,
                        out_channels=self.num_output_channels if i == self.num_upsampling_layers - 1 else self.num_channels,
                        kernel_size=4,
                        stride=2,
                        padding=1,
                        use_norm=use_norm
                    ),
                    ActivationCls() if i != self.num_upsampling_layers - 1 else nn.Identity(),
                ]
            )
        self.blocks = nn.Sequential(*blocks)

    def forward(self, x):
        return self.blocks(x)
