import torch
import torch.nn as nn
from einops import rearrange

from utils import is_power_of_two
from vqa.einops_autoencoder import Encoder, Decoder, Conv3dLayer

from vqa.vector_quantizer import VectorQuantizer


class FlattenAndRestructure(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self,
                x: torch.FloatTensor    # [N, C, D, H, W]
                ):
        d, h, w = x.shape[-3:]
        x = rearrange(x, 'n c d h w -> n (d h w) c')
        x_quantized, x_discrete, latent_loss = self.fn(x)
        x_quantized = rearrange(x_quantized, 'n (d h w) c -> n c d h w', d=d, h=h, w=w)
        x_discrete = rearrange(x_discrete, 'n (d h w) -> n d h w', d=d, h=h, w=w)
        return x_quantized, x_discrete, latent_loss

    def discrete_to_codebook_vectors(self, discrete):
        return rearrange(self.fn.get_codes_from_indices(discrete), 'n h w d e -> n e h w d')


class Autoencoder(nn.Module):
    def __init__(
            self,
            img_size: int,
            img_size_top: int,
            num_layers_per_block: int,
            num_embedding_channels_top: int,
            num_codebook_vectors_top: int,
            codebook_dim_top: int,
            vector_quantizer_kwargs,
            activation: str,
            kernel_size: int,
            use_norm: bool,
            num_input_channels: int,
            is_training: bool,
            **kwargs,
    ):
        assert is_power_of_two(img_size)
        assert is_power_of_two(img_size_top)
        assert img_size % img_size_top == 0
        super().__init__()

        self.img_size = img_size
        self.img_size_top = img_size_top
        self.kernel_size = kernel_size
        self.num_layers_per_block = num_layers_per_block
        self.num_codebook_vectors_top = num_codebook_vectors_top
        self.use_norm = use_norm
        self.num_input_channels = num_input_channels
        self.num_embedding_channels_top = num_embedding_channels_top
        self.codebook_dim_top = codebook_dim_top
        self.activation = activation
        self.is_training = is_training

        self.encoder = Encoder(
            input_size=self.img_size,
            output_size=self.img_size_top,
            kernel_size=self.kernel_size,
            num_input_channels=self.num_input_channels,
            num_channels=self.num_embedding_channels_top,
            num_layers_per_block=self.num_layers_per_block,
            use_norm=self.use_norm,
            stride=2,
            activation=self.activation
        )

        self.vector_quantizer_conv_top = nn.ModuleDict(
            {
                "in": Conv3dLayer(in_channels=self.num_embedding_channels_top, out_channels=self.codebook_dim_top, kernel_size=1, use_norm=self.use_norm),
                "out": Conv3dLayer(in_channels=self.codebook_dim_top, out_channels=self.num_embedding_channels_top, kernel_size=1, use_norm=self.use_norm),
            }
        )

        self.vector_quantizer = FlattenAndRestructure(VectorQuantizer(num_codebook_vectors=self.num_codebook_vectors_top,
                                                                      dim=self.codebook_dim_top,
                                                                      codebook_dim=self.codebook_dim_top,
                                                                      is_training=self.is_training,
                                                                      **vector_quantizer_kwargs))

        self.decoder = Decoder(
            input_size=self.img_size,
            output_size=self.img_size_top,
            kernel_size=self.kernel_size,
            num_output_channels=self.num_input_channels,
            num_channels=self.num_embedding_channels_top,
            use_norm=self.use_norm,
            num_layers_per_block=self.num_layers_per_block,
            stride=2,
            activation=self.activation
        )

    def discrete_to_codebook_vectors(self, discrete: torch.LongTensor):
        return self.vector_quantizer.discrete_to_codebook_vectors(discrete)

    def encode_discrete(self, x: torch.FloatTensor):
        encoded_top = self.encoder(x)

        pre_quantized_top = self.vector_quantizer_conv_top["in"](encoded_top)
        quantized_top, discrete_top, latent_loss_top = self.vector_quantizer(pre_quantized_top)

        return quantized_top, latent_loss_top, discrete_top

    def decode_discrete(self, quantized_top):
        post_quantized_top = self.vector_quantizer_conv_top["out"](quantized_top)
        decoded = self.decoder(post_quantized_top)
        return decoded

    def encode_continuous(self, x: torch.FloatTensor):
        encoded_top = self.encoder(x)
        return encoded_top

    def decode_continuous(self, encoded_top):
        pre_quantized_top = self.vector_quantizer_conv_top["in"](encoded_top)
        quantized_top, discrete_top, latent_loss_top = self.vector_quantizer(pre_quantized_top)

        post_quantized_top = self.vector_quantizer_conv_top["out"](quantized_top)
        decoded = self.decoder(post_quantized_top)

        return decoded, latent_loss_top, discrete_top

    def forward(self, x: torch.FloatTensor):
        quantized_top, latent_loss, discrete_top = self.encode_discrete(x)
        decoded = self.decode_discrete(quantized_top)
        return decoded, latent_loss, discrete_top
