import shutil
import random
import os
import numpy as np
import sys
from argparse import ArgumentParser
import pathlib
import torchio
from tempfile import gettempdir

from tqdm import tqdm

from easydict import EasyDict
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import wandb
from einops import reduce
from torchinfo import summary

import oyaml as yaml

from vqa.einops_vqa import Autoencoder
from data import MriDataLoaderTorchIO, tensor_to_image_grid

from accelerate import Accelerator, DistributedDataParallelKwargs
from utils import visualize_mri_slices
from lion_pytorch import Lion

from utils import strip_distributed_markers

from einops._torch_specific import allow_ops_in_compiled_graph  # requires einops>=0.6.1

allow_ops_in_compiled_graph()


def seed_everything(seed: int):
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def cycle(dl):
    while True:
        for data in dl:
            yield data


def main(config_fname: pathlib.Path, dry_run: bool, load_ckpt: bool, tmp_dir: pathlib.Path):
    # Config
    with open(config_fname, "r") as config_f:
        autoencoder_config = EasyDict(yaml.load(config_f, Loader=yaml.FullLoader))

    # Accelerator
    ddp_kwargs = DistributedDataParallelKwargs(find_unused_parameters=autoencoder_config.mixed_precision != "no")
    accelerator = Accelerator(mixed_precision=autoencoder_config.mixed_precision, kwargs_handlers=[ddp_kwargs], gradient_accumulation_steps=autoencoder_config.num_accumulation_steps)
    tmp_dir = tmp_dir.joinpath(f"Autoencoder_3d_{autoencoder_config.img_size}")

    if accelerator.is_local_main_process:
        original_size = 3 * autoencoder_config.img_size**3

        top_size_continuous = autoencoder_config.autoencoder.num_embedding_channels_top * autoencoder_config.img_size_top**3
        compression_factor_continuous = 100.0 * top_size_continuous / original_size
        print(f"Compression factor [continuous]: {compression_factor_continuous:.3f}%")

        top_size_discrete = autoencoder_config.img_size_top**3
        compression_factor_discrete = 100.0 * top_size_discrete / original_size
        print(f"Compression factor [discrete]: {compression_factor_discrete:.3f}%")

    if accelerator.is_local_main_process and not args.dry_run:

        tmp_dir.mkdir(exist_ok=True)
        shutil.copyfile(src=config_fname, dst=tmp_dir.joinpath(f"config_autoencoder_{autoencoder_config.data_mode}.yml"))

        # Logging
        wandb.init(project="<PROJECT>", entity="<ENTITY>")
        wandb.run.log_code(".")
        wandb.define_metric("train/global_step")
        wandb.define_metric("train/*", step_metric="train/global_step")
        wandb.config.update(autoencoder_config)

    autoencoder_ckpt_fname = tmp_dir.joinpath(f"autoencoder_{autoencoder_config.data_mode}.pt")

    # Dataset
    data = MriDataLoaderTorchIO(
        dataset_dir=pathlib.Path(autoencoder_config.dataset_dir),
        index_fname=pathlib.Path(autoencoder_config.index_fname),
        prefix=autoencoder_config.prefix,
        variant=autoencoder_config.variant,
        data_mode=autoencoder_config.data_mode,
        img_size=autoencoder_config.img_size,
        augmentation_kwargs=autoencoder_config.augmentation_kwargs,
        batch_size=autoencoder_config.batch_size,
        num_workers=8,
    )

    autoencoder = Autoencoder(
        num_input_channels=1,
        img_size=autoencoder_config.img_size,
        img_size_top=autoencoder_config.img_size_top,
        is_training=True,
        **autoencoder_config.autoencoder,
    )

    if load_ckpt:
        ckpt = torch.load(autoencoder_ckpt_fname, map_location="cpu")
        ckpt = strip_distributed_markers(ckpt, pattern="_orig_mod.module.", replacement="")
        autoencoder.load_state_dict(ckpt)

    OptimizerCls = Lion if autoencoder_config.optimizer.pop("type") == "Lion" else optim.Adam
    optimizer = OptimizerCls(autoencoder.parameters(), **autoencoder_config.optimizer)
    loss_fct = nn.MSELoss()

    autoencoder, optimizer, data = accelerator.prepare(autoencoder, optimizer, data)

    data = cycle(data)
    global_step = 0
    for batch in tqdm(data, disable=not accelerator.is_local_main_process):
        batch = EasyDict(voxel_values_unaug=batch["voxel_values_unaug"][torchio.DATA], voxel_values_aug=batch["voxel_values_aug"][torchio.DATA], stems=batch["voxel_values_unaug"]["stem"])

        summary(autoencoder, input_data=batch.voxel_values_aug)
        batch.voxel_values_aug_decoded, latent_loss, batch.discrete_top = autoencoder(batch.voxel_values_aug)

        target = None
        if autoencoder_config.compute_loss_between == "aug_unaug":
            target = batch.voxel_values_unaug
        elif autoencoder_config.compute_loss_between == "aug_aug":
            target = batch.voxel_values_aug
        else:
            raise ValueError("Unknown loss mode!")

        reconstruction_loss = loss_fct(batch.voxel_values_aug_decoded, target)
        loss = latent_loss + reconstruction_loss

        loss.backward()
        optimizer.step()
        optimizer.zero_grad()

        if dry_run or not accelerator.is_local_main_process:
            global_step += 1
            continue

        with torch.no_grad():

            if global_step != 0 and global_step % autoencoder_config.log_step == 0:

                discrete_top_ohe = F.one_hot(batch.discrete_top, autoencoder_config.autoencoder.num_codebook_vectors_top)
                top_nnz = 100.0 * (reduce(discrete_top_ohe, "n h w d c -> c", "sum") > 0).float().mean()

                log_dict = {
                    "train/global_step": global_step,
                    "train/loss": loss,
                    "train/reconstruction_loss": reconstruction_loss,
                    "train/latent_loss": latent_loss,
                    "train/nnz_percentage_top": top_nnz,
                }

                if global_step % autoencoder_config.visualization_step == 0:

                    if autoencoder_config.data_mode == "brain_mask":
                        batch.voxel_values_aug_decoded.sigmoid_()

                    slices_unaug = visualize_mri_slices(batch.voxel_values_unaug, type="axial", min_max_scaling=True)
                    slices_aug = visualize_mri_slices(batch.voxel_values_aug, type="axial", min_max_scaling=True)
                    slices_aug_decoded = visualize_mri_slices(batch.voxel_values_aug_decoded, type="axial", min_max_scaling=autoencoder_config.data_mode == "mri")

                    log_img_dict = {
                        "train/slices_aug": wandb.Image(tensor_to_image_grid(slices_aug, shift=False)),
                        "train/slices_unaug": wandb.Image(tensor_to_image_grid(slices_unaug, shift=False)),
                        "train/slices_decoded": wandb.Image(tensor_to_image_grid(slices_aug_decoded, shift=False)),
                    }
                    log_dict = {**log_dict, **log_img_dict}

                wandb.log(log_dict)

                if torch.isnan(batch.voxel_values_aug_decoded).any():
                    print("Aborting training due to NaNs")
                    sys.exit(1)

            if not args.dry_run and (global_step + 1) % autoencoder_config.save_step == 0:
                torch.save(autoencoder.state_dict(), autoencoder_ckpt_fname)

        global_step += 1


if __name__ == "__main__":
    seed_everything(42)
    parser = ArgumentParser()

    parser.add_argument("--config_fname", required=True)

    parser.add_argument("--dry_run", action="store_true", default=False)

    parser.add_argument("--load_ckpt", action="store_true", default=False)

    parser.add_argument("--tmp_dir", default=gettempdir())

    args = parser.parse_args()

    args.config_fname = pathlib.Path(args.config_fname)
    assert args.config_fname.exists(), "Config does not exist!"

    args.tmp_dir = pathlib.Path(args.tmp_dir)

    main(config_fname=args.config_fname, dry_run=args.dry_run, load_ckpt=args.load_ckpt, tmp_dir=args.tmp_dir)
